#include "Player.hpp"

CGUE17_SneakySteve::scene::Player::Player(std::string path, sneakystv::Shader * _shader, glm::vec3 physicsCorr, float xRot, float yRot, float zRot, btRigidBody * rigidBody, Camera* camera)
	:GameObj(path, _shader, physicsCorr, xRot, yRot, zRot, rigidBody), camera(camera)
{
}

CGUE17_SneakySteve::scene::Player::~Player()
{
}

int CGUE17_SneakySteve::scene::Player::draw(Shader* shader, glm::mat4 projview)
{
	glm::mat4 modelMat = glm::mat4(1);

	//add translation to modelMat
	modelMat = glm::translate(modelMat, position);

	//add Rotation to modelMat
	modelMat = glm::rotate(modelMat, xRot, glm::vec3(-1, 0, 0));
	modelMat = glm::rotate(modelMat, yRot, glm::vec3(0, -1, 0));
	modelMat = glm::rotate(modelMat, zRot, glm::vec3(0, 0, 1));

	return sceneObj->draw(shader, modelMat, glm::vec3(0, 2, 0), projview);
}

void CGUE17_SneakySteve::scene::Player::update(float time_delta)
{

	btTransform trans;
	rigidBody->getMotionState()->getWorldTransform(trans);
	
	//get position of rigidbody
	position = glm::vec3(trans.getOrigin().getX(), trans.getOrigin().getY(), trans.getOrigin().getZ()) + physicsCorr;

	//set Camera Position
	glm::vec3 headpos = glm::vec3(position.x, position.y + 2, position.z);
	
	glm::vec3 camPos = headpos - (camera->getCamFront() * 10.0f);



	camera->position = camPos;




	// Still makes Problems no idea why tho
	/*btQuaternion quat;
	quat.setEulerZYX(zRot, yRot, xRot); //or quat.setEulerZYX depending on the ordering you want
	trans.setRotation(quat);
	
	rigidBody->setCenterOfMassTransform(trans);*/

	//std::cout << "sphere height: " << trans.getOrigin().getY() << std::endl;
}

glm::vec3 CGUE17_SneakySteve::scene::Player::getFront()
{
	glm::vec3 front;
	front.x = cos(xRot) * cos(yRot);
	front.y = sin(xRot);
	front.z = cos(xRot) * sin(yRot);

	front = glm::normalize(front);
	return front;
}

void CGUE17_SneakySteve::scene::Player::translate(float x, float y, float z)
{
	if (x != 0.0 || y != 0.0 || z != 0.0)
	{
		yRot = camera->yaw;
	}

	x *= 1;
	y *= 1;
	z *= 1;

	glm::vec3 front;
	front.x = cos(xRot) * cos(yRot);
	front.y = sin(xRot);
	front.z = cos(xRot) * sin(yRot);

	front = glm::normalize(front);

	glm::vec3 velocity;

	velocity = front * x;
	velocity += glm::normalize(glm::cross(front, glm::vec3(0.0f, -1.0f, 0.0f)))*y;
	if(velocity.x > 1 ||velocity.y > 1|| velocity.z > 1)
		printf("xRot: %f| yRot: %f| zRot: %f| velx: %f| vely: %f| velz: %f  \n",xRot,yRot,zRot,velocity.x,velocity.y,velocity.z );


	btVector3 btcurrVel = rigidBody->getLinearVelocity();
	btVector3 btvelocity = btVector3(velocity.x, btcurrVel.getY(), velocity.z);

	rigidBody->setLinearVelocity(btvelocity);
	
}

void CGUE17_SneakySteve::scene::Player::rotate(double x, double y)
{
	

	if (y == 0)
		return;

	
	yRot += y;

	//Rotate Solid body here;

}

void CGUE17_SneakySteve::scene::Player::jump()
{

	btVector3 btcurrVel = rigidBody->getLinearVelocity();
	if (btcurrVel.getY() <0.1 && btcurrVel.getY() > -0.1)
	{
		//Jumping (speed)

		/*btVector3 btvelocity = btVector3(btcurrVel.getX(),5, btcurrVel.getZ());
		rigidBody->setLinearFactor(btvelocity);*/



		//Jumping (translation)

		btTransform trans;
		rigidBody->getMotionState()->getWorldTransform(trans);
		trans.setOrigin(btVector3(trans.getOrigin().getX(), trans.getOrigin().getY() + 2, trans.getOrigin().getZ()));
		trans.setRotation(btQuaternion(0, 0, 0, 1));
		rigidBody->setWorldTransform(trans);	

	}
	
}
