#include "Warden.h"

CGUE17_SneakySteve::scene::Warden::Warden(std::string path, sneakystv::Shader * _shader, glm::vec3 physicsCorr, float xRot, float yRot, float zRot, btRigidBody * rigidBody, std::vector<glm::vec2> Waypoints, Camera* camera)
	:GameObj(path, _shader, physicsCorr, xRot, yRot, zRot, rigidBody), waypts(Waypoints), currWayptsIdx(0)
{
	id = 1;
	particle = new ParticleSpawner(Shaderholder::getShader(SHADER_PARTICLE), camera);
}

CGUE17_SneakySteve::scene::Warden::~Warden()
{
	//Dont forget
	id = 1;
}

void CGUE17_SneakySteve::scene::Warden::update(float time_delta)
{
	

	int speed = 15;
	
	btTransform trans;
	rigidBody->getMotionState()->getWorldTransform(trans);

	//particle Spawner update
	particle->update(time_delta);
	particle->position = glm::vec3(trans.getOrigin().getX(), trans.getOrigin().getY(), trans.getOrigin().getZ());
	
	position = glm::vec3(trans.getOrigin().getX(), trans.getOrigin().getY(), trans.getOrigin().getZ());

	glm::vec2 position = glm::vec2(trans.getOrigin().getX(), trans.getOrigin().getZ());
	glm::vec2 currTarget = waypts.at(currWayptsIdx);


	//Check if Warden has reached his target Waypoint
	if (glm::length(currTarget-position)< 1) {
		currWayptsIdx++;
		if (currWayptsIdx > waypts.size() - 1)
			currWayptsIdx = 0;
	}

	currTarget = waypts.at(currWayptsIdx);

	glm::vec2 direction = (currTarget - position);
	direction = glm::normalize(direction);
	direction *= speed;
	//direction *= time_delta;

	btVector3 currVelocity = rigidBody->getLinearVelocity();
	btVector3 velocity = btVector3(direction.x, currVelocity.getY(), direction.y);
	
	btQuaternion rotation = trans.getRotation();

	rigidBody->setLinearVelocity(velocity);

}

int CGUE17_SneakySteve::scene::Warden::draw(Shader* shader, glm::mat4 projview)
{
	glm::mat4 modelMat;
	rigidBody->getWorldTransform().getOpenGLMatrix(glm::value_ptr(modelMat));
	modelMat = glm::translate(modelMat, physicsCorr);
	

	if (shader != Shaderholder::getShader(SHADER_SHADOW)) 
	{
		particle->Draw();
		shader->useShader();
	}

	return sceneObj->draw(shader, modelMat, glm::vec3(1, 0, 0), projview);
}

int CGUE17_SneakySteve::scene::Warden::getId()
{
	return id;
}
