#pragma once

#include "PhysicsManager.hpp"
#include <BulletCollision\Gimpact\btGImpactCollisionAlgorithm.h>

PhysicsManager::PhysicsManager()
{
	broadphase = new btDbvtBroadphase();
	collisionConfiguration = new btDefaultCollisionConfiguration();
	dispatcher = new btCollisionDispatcher(collisionConfiguration);
	btGImpactCollisionAlgorithm::registerAlgorithm(dispatcher);
	solver = new btSequentialImpulseConstraintSolver;
	dynamicsWorld = new btDiscreteDynamicsWorld(dispatcher, broadphase, solver, collisionConfiguration);
	dynamicsWorld->setGravity(btVector3(0, -10, 0));

	groundShape = new btStaticPlaneShape(btVector3(0, 1, 0), 1);
	fallShape = new btSphereShape(1);

	groundMotionState = new btDefaultMotionState(btTransform(btQuaternion(0, 0, 0, 1), btVector3(0, -1, 0)));
	btRigidBody::btRigidBodyConstructionInfo
		groundRigidBodyCI(0, groundMotionState, groundShape, btVector3(0, 0, 0));
	groundRigidBody = new btRigidBody(groundRigidBodyCI);
	dynamicsWorld->addRigidBody(groundRigidBody);

	fallMotionState =
		new btDefaultMotionState(btTransform(btQuaternion(0, 0, 0, 1), btVector3(0, 50, 0)));
	mass = 1;
	btVector3 fallInertia(0, 0, 0);
	fallShape->calculateLocalInertia(mass, fallInertia);
	btRigidBody::btRigidBodyConstructionInfo fallRigidBodyCI(mass, fallMotionState, fallShape, fallInertia);
	fallRigidBody = new btRigidBody(fallRigidBodyCI);
	dynamicsWorld->addRigidBody(fallRigidBody);
}

void PhysicsManager::addPhysicsObject(char type, btScalar transform[], btScalar mass, glm::vec3 inertia) {
	btCollisionShape* collisionShape;
	btVector3 min;
	btVector3 max;
	switch (type)
	{
	case 'b': collisionShape = new btBoxShape(btVector3(0.1, 0.1, 0.1));
		break;
	case 'p': collisionShape = new btStaticPlaneShape(btVector3(0, 1, 0), 1);
		break;
	case 'c': collisionShape = new btSphereShape(2);
		break;
	default:
		break;
	}
	collisionShapes.push_back(collisionShape); 

	btVector3 localInertia = btVector3(inertia.x, inertia.y, inertia.z);
	collisionShape->calculateLocalInertia(mass, localInertia);
	btTransform localTransform = btTransform();
	localTransform.setFromOpenGLMatrix(transform);
	btDefaultMotionState* motionState = new btDefaultMotionState(localTransform);
	btRigidBody::btRigidBodyConstructionInfo
		rigidBodyCI(mass, motionState, collisionShape, localInertia);
	btRigidBody* rigidBody;
	if (type == 'c') {
		rigidBodyCI.m_linearDamping = 0.9f;
		rigidBody = new btRigidBody(rigidBodyCI);
		playerRigidBody = rigidBody;
		playerRigidBody->setAngularFactor(btVector3(0, 1, 0));
		playerRigidBody->setActivationState(DISABLE_DEACTIVATION);

	}
	else if (type == 'p' || type == 'b') {
		rigidBody = new btRigidBody(rigidBodyCI);
		rigidBody->setFlags(BT_DISABLE_WORLD_GRAVITY);
		rigidBodys.push_back(rigidBody);
	}
	else {
		rigidBody = new btRigidBody(rigidBodyCI);
		//rigidBodys.push_back(rigidBody);
	}

	dynamicsWorld->addRigidBody(rigidBody);
}

glm::vec3 PhysicsManager::movePlayer(glm::vec3 cameraPosition, /*glm::vec3 playerPos,*/ float time_delta) {
	//playerPos = glm::vec3(playerRigidBody->getCenterOfMassPosition().x(), playerRigidBody->getCenterOfMassPosition().y(), playerRigidBody->getCenterOfMassPosition().z());
	playerRigidBody->applyCentralForce(btVector3(500 / time_delta * (cameraPosition.x /*- playerPos.x*/), 10 / time_delta * (cameraPosition.y /*- playerPos.y*/), 500 / time_delta * (cameraPosition.z/* - playerPos.z*/)));
	playerRigidBody->applyDamping(0);

	btTransform playerTrans;
	GLfloat playerPositionPointer[16];
	dynamicsWorld->stepSimulation(time_delta, 10);
	playerRigidBody->getMotionState()->getWorldTransform(playerTrans);
	playerTrans.getOpenGLMatrix(playerPositionPointer);
	return glm::vec3(playerPositionPointer[12], playerPositionPointer[13], playerPositionPointer[14]);
}

PhysicsManager::~PhysicsManager()
{
	for (auto i = dynamicsWorld->getNumCollisionObjects() - 1; i >= 0; i--)
	{
		btCollisionObject* object = dynamicsWorld->getCollisionObjectArray()[i];
		btRigidBody* rigidBody = btRigidBody::upcast(object);
		if (rigidBody && rigidBody->getMotionState())
		{
			delete rigidBody->getMotionState();
		}
		dynamicsWorld->removeCollisionObject(object);
		delete object;
	}

	for (auto i = 0; i < collisionShapes.size(); i++)
	{
		btCollisionShape* shape = collisionShapes[i];
		collisionShapes[i] = 0;
		delete shape;
	}

	delete dynamicsWorld;
	delete solver;
	delete dispatcher;
}