#include "Mesh.hpp"
#include <vector>
#include <fstream>
#include <sstream>
#include <iostream>

#include "..\Util\KeyMapping.hpp"

using namespace CGUE17_SneakySteve::scene;
using namespace std;

Mesh::Mesh(vector<Vertex> vertices, vector<GLuint> indices, vector<Texture> textures)
{
	this->vertices = vertices;
	this->indices = indices;
	this->textures = textures;


	for (int i = 0; i < vertices.size(); i++)
	{
		if (i == 0)
		{
			minAABB = { vertices[i].Normal.x, vertices[i].Normal.y , vertices[i].Normal.z };
			maxAABB = { vertices[i].Normal.x, vertices[i].Normal.y , vertices[i].Normal.z };
		}
		minAABB = {
			minAABB.x<vertices[i].Position.x ? minAABB.x : vertices[i].Position.x,
			minAABB.y<vertices[i].Position.y ? minAABB.y : vertices[i].Position.y,
			minAABB.z<vertices[i].Position.z ? minAABB.z : vertices[i].Position.z
		};
		maxAABB = {
			maxAABB.x>vertices[i].Position.x ? maxAABB.x : vertices[i].Position.x,
			maxAABB.y>vertices[i].Position.y ? maxAABB.y : vertices[i].Position.y,
			maxAABB.z>vertices[i].Position.z ? maxAABB.z : vertices[i].Position.z
		};
	}

	this->setupMesh();
}



void Mesh::setupMesh()
{
	glGenVertexArrays(1, &this->VAO);
	glGenBuffers(1, &this->VBO);
	glGenBuffers(1, &this->EBO);

	glBindVertexArray(this->VAO);
	glBindBuffer(GL_ARRAY_BUFFER, this->VBO);

	glBufferData(GL_ARRAY_BUFFER, this->vertices.size() * sizeof(Vertex),
		&this->vertices[0], GL_STATIC_DRAW);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, this->EBO);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, this->indices.size() * sizeof(GLuint),
		&this->indices[0], GL_STATIC_DRAW);

	// Vertex Positions
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex),
		(GLvoid*)0);
	// Vertex Normals
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex),
		(GLvoid*)offsetof(Vertex, Normal));
	// Vertex Texture Coords
	glEnableVertexAttribArray(2);
	glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex),
		(GLvoid*)offsetof(Vertex, TexCoords));

	glBindVertexArray(0);
}

void Mesh::Draw(sneakystv::Shader* shader)
{
	GLuint diffuseNr = 1;
	GLuint specularNr = 1;
	for (GLuint i = 0; i < this->textures.size(); i++)
	{
		stringstream ss;
		string number;
		string name = this->textures[i].type;
		if (name == "texture_diffuse")
			ss << diffuseNr++; // Transfer GLuint to stream
		else if (name == "texture_specular")
			ss << specularNr++; // Transfer GLuint to stream
		number = ss.str();
		glUniform1f(glGetUniformLocation(shader->programHandle, ("material." + name + number).c_str()), i);
		glActiveTexture(GL_TEXTURE0 + i); // Activate proper texture unit before binding
										  // Retrieve texture number (the N in diffuse_textureN)
		if(KeyMapping::getWireFrame())
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		else
			glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
		
		glBindTexture(GL_TEXTURE_2D, this->textures[i].id);
	}
	glActiveTexture(GL_TEXTURE0);

	// Draw mesh
	
	
	glBindVertexArray(this->VAO);
	glDrawElements(GL_TRIANGLES, this->indices.size(), GL_UNSIGNED_INT, 0);
	glBindVertexArray(0);
}