#include "Model.hpp"
#include <assimp\Logger.hpp>
#include <assimp\DefaultLogger.hpp>
#include <stdbool.h>
#include <iostream>

#include<soil\SOIL.h>

#include "..\Util\FrustumCulling.hpp"
#include "glm\common.hpp"
#include "..\Util\KeyMapping.hpp"

using namespace CGUE17_SneakySteve::scene;

Model::Model(string/*GLchar**/ path, Shader* shader)
{
	this->loadModel(path, shader);
}

int Model::Draw(Shader* shader, glm::mat4 modelmat, glm::vec3 color, glm::mat4 projview)
{
	auto model_location = glGetUniformLocation(shader->programHandle, "model");
	glUniformMatrix4fv(model_location, 1, GL_FALSE, glm::value_ptr(modelmat));

	model_location = glGetUniformLocation(shader->programHandle, "color");
	glUniform3f(model_location, color.x, color.y, color.z);

	FrustumCulling::computePlanes((float*)glm::value_ptr(projview*modelmat));
	int culled = 0;
	bool inside = true;
	if (KeyMapping::getFrCulling())
	{
		for (int i = 0; i < 6; ++i)
		{
			int out = 0;
			out += (glm::dot(FrustumCulling::planes[i]->normal, glm::vec3(minAABB.x, minAABB.y, minAABB.z)) + FrustumCulling::planes[i]->d < 0) ? 1 : 0;
			out += (glm::dot(FrustumCulling::planes[i]->normal, glm::vec3(maxAABB.x, minAABB.y, minAABB.z)) + FrustumCulling::planes[i]->d < 0) ? 1 : 0;
			out += (glm::dot(FrustumCulling::planes[i]->normal, glm::vec3(minAABB.x, maxAABB.y, minAABB.z)) + FrustumCulling::planes[i]->d < 0) ? 1 : 0;
			out += (glm::dot(FrustumCulling::planes[i]->normal, glm::vec3(maxAABB.x, maxAABB.y, minAABB.z)) + FrustumCulling::planes[i]->d < 0) ? 1 : 0;
			out += (glm::dot(FrustumCulling::planes[i]->normal, glm::vec3(minAABB.x, minAABB.y, maxAABB.z)) + FrustumCulling::planes[i]->d < 0) ? 1 : 0;
			out += (glm::dot(FrustumCulling::planes[i]->normal, glm::vec3(maxAABB.x, minAABB.y, maxAABB.z)) + FrustumCulling::planes[i]->d < 0) ? 1 : 0;
			out += (glm::dot(FrustumCulling::planes[i]->normal, glm::vec3(minAABB.x, maxAABB.y, maxAABB.z)) + FrustumCulling::planes[i]->d < 0) ? 1 : 0;
			out += (glm::dot(FrustumCulling::planes[i]->normal, glm::vec3(maxAABB.x, maxAABB.y, maxAABB.z)) + FrustumCulling::planes[i]->d < 0) ? 1 : 0;
			//std::cout << out << std::endl;
			if (out == 8) inside = false;
		}
	}
	//culled = (KeyMapping::getFrCulling() ? (frustumTestBox(projview, minAABB, maxAABB) ? 1 : 0) : 0);
		if (/*culled == 0 */inside)
		{
			for (GLuint i = 0; i < this->meshes.size(); i++)
			{
				this->meshes[i].Draw(shader);
			}
		}
		else
		{
			culled++;

		}
	//std::cout << culled << std::endl;
	return culled;
}

bool CGUE17_SneakySteve::scene::Model::frustumTestBox(glm::mat4 mvp, glm::vec3 minAABB, glm::vec3 maxAABB)
{
	if (!KeyMapping::getFrCulling()) return false;

	

	glm::vec4 bbox[8];

	bbox[0] = glm::vec4{ minAABB, 1.0f };
	bbox[1] = glm::vec4{ minAABB.x, minAABB.y, maxAABB.z, 1.0f };
	bbox[2] = glm::vec4{ minAABB.x, maxAABB.y, minAABB.z, 1.0f };
	bbox[3] = glm::vec4{ maxAABB.x, minAABB.y, minAABB.z, 1.0f };

	bbox[4] = glm::vec4{ maxAABB, 1.0f };
	bbox[5] = glm::vec4{ maxAABB.x, maxAABB.y, minAABB.z, 1.0f };
	bbox[6] = glm::vec4{ maxAABB.x, minAABB.y, maxAABB.z, 1.0f };
	bbox[7] = glm::vec4{ minAABB.x, maxAABB.y, maxAABB.z, 1.0f };

	int planes_cnt[6] = { 0, 0, 0, 0, 0, 0 };

	for (int i = 0; i < 8; i++) {
		glm::vec4 cur = mvp * bbox[i];
		if (cur.x >  cur.w) planes_cnt[0]++;
		if (cur.x < -cur.w) planes_cnt[1]++;
		if (cur.y >  cur.w) planes_cnt[2]++;
		if (cur.y < -cur.w) planes_cnt[3]++;
		if (cur.z >  cur.w) planes_cnt[4]++;
		if (cur.z < -cur.w) planes_cnt[5]++;
	}

	for (int i = 0; i < 6; i++) {
		// Outside the frustum
		if (planes_cnt[i] == 8) {
			return true;
		}
	}
	// Inside the frustum
	return false;
}

bool Model::loadModel(string path, Shader* shader)
{
	Assimp::Importer importer;
	const aiScene* scene = importer.ReadFile(path, aiProcess_Triangulate | aiProcess_FlipUVs);
	
	//Assimp::DefaultLogger::create("", Assimp::Logger::VERBOSE);

	if (!scene)
	{
		//Assimp::DefaultLogger::get()->error(importer.GetErrorString);
		return false;
	}

	this->directory = path.substr(0, path.find_last_of('/'));

	this->processNode(scene->mRootNode, scene,shader);
	return true;
}

void Model::processNode(aiNode* node, const aiScene* scene, Shader* shader)
{
	// Process all the node's meshes (if any)
	for (GLuint i = 0; i < node->mNumMeshes; i++)
	{
		aiMesh* mesh = scene->mMeshes[node->mMeshes[i]];
		Mesh forAABB = this->processMesh(mesh, scene, shader);
		this->meshes.push_back(forAABB);

		if (i == 0)
		{
			minAABB = { forAABB.minAABB.x, forAABB.minAABB.y, forAABB.minAABB.z };
			maxAABB = { forAABB.maxAABB.x, forAABB.maxAABB.y, forAABB.maxAABB.z };
		}

		minAABB = {
			minAABB.x < forAABB.minAABB.x ? minAABB.x : forAABB.minAABB.x,
			minAABB.y < forAABB.minAABB.y ? minAABB.y : forAABB.minAABB.y,
			minAABB.z < forAABB.minAABB.z ? minAABB.z : forAABB.minAABB.z
		};
		maxAABB = {
			maxAABB.x > forAABB.maxAABB.x ? maxAABB.x : forAABB.maxAABB.x,
			maxAABB.y > forAABB.maxAABB.y ? maxAABB.y : forAABB.maxAABB.y,
			maxAABB.z > forAABB.maxAABB.z ? maxAABB.z : forAABB.maxAABB.z
		};
	}
	// Then do the same for each of its children
	for (GLuint i = 0; i < node->mNumChildren; i++)
	{
		this->processNode(node->mChildren[i],scene,shader);
	}
}

Mesh Model::processMesh(aiMesh* mesh, const aiScene* scene, Shader* shader)
{
	// Data to fill
	vector<Vertex> vertices;
	vector<GLuint> indices;
	vector<Texture> textures;

	// Walk through each of the mesh's vertices
	for (GLuint i = 0; i < mesh->mNumVertices; i++)
	{
		Vertex vertex;
		glm::vec3 vector; 
		// We declare a placeholder vector since assimp uses its own vector class that doesn't directly convert to glm's vec3 class so we transfer the data to this placeholder glm::vec3 first.
						  
		// Positions
		vector.x = mesh->mVertices[i].x;
		vector.y = mesh->mVertices[i].y;
		vector.z = mesh->mVertices[i].z;
		vertex.Position = vector;
		// Normals
		vector.x = mesh->mNormals[i].x;
		vector.y = mesh->mNormals[i].y;
		vector.z = mesh->mNormals[i].z;
		vertex.Normal = vector;
		// Texture Coordinates
		if (mesh->mTextureCoords[0]) // Does the mesh contain texture coordinates?
		{
			glm::vec2 vec;
			// A vertex can contain up to 8 different texture coordinates. We thus make the assumption that we won't 
			// use models where a vertex can have multiple texture coordinates so we always take the first set (0).
			vec.x = mesh->mTextureCoords[0][i].x;
			vec.y = mesh->mTextureCoords[0][i].y;
			vertex.TexCoords = vec;
		}
		else
			vertex.TexCoords = glm::vec2(0.0f, 0.0f);
		vertices.push_back(vertex);
	}
	// Now wak through each of the mesh's faces (a face is a mesh its triangle) and retrieve the corresponding vertex indices.
	for (GLuint i = 0; i < mesh->mNumFaces; i++)
	{
		aiFace face = mesh->mFaces[i];
		// Retrieve all indices of the face and store them in the indices vector
		for (GLuint j = 0; j < face.mNumIndices; j++)
			indices.push_back(face.mIndices[j]);
	}
	// Process materials
	if (mesh->mMaterialIndex >= 0)
	{
		aiMaterial* material = scene->mMaterials[mesh->mMaterialIndex];
		// We assume a convention for sampler names in the shaders. Each diffuse texture should be named
		// as 'texture_diffuseN' where N is a sequential number ranging from 1 to MAX_SAMPLER_NUMBER. 
		// Same applies to other texture as the following list summarizes:
		// Diffuse: texture_diffuseN
		// Specular: texture_specularN
		// Normal: texture_normalN

		// 1. Diffuse maps
		vector<Texture> diffuseMaps = this->loadMaterialTextures(material, aiTextureType_DIFFUSE, "texture_diffuse", shader );
		textures.insert(textures.end(), diffuseMaps.begin(), diffuseMaps.end());
		// 2. Specular maps
		vector<Texture> specularMaps = this->loadMaterialTextures(material, aiTextureType_SPECULAR, "texture_specular", shader);
		textures.insert(textures.end(), specularMaps.begin(), specularMaps.end());
	}

	// Return a mesh object created from the extracted mesh data
	return Mesh(vertices, indices, textures);
}


std::vector<Texture> Model::loadMaterialTextures(aiMaterial* material, aiTextureType textureType, std::string typeName, Shader* shader)
{
	std::vector<Texture> textures;
	for (size_t i = 0; i < material->GetTextureCount(textureType); i++)
	{
		aiString string;
		material->GetTexture(textureType, i, &string);
		string = directory + '/' + string.C_Str();
		std::string str = string.C_Str();
		std::string wide(str.begin(), str.end());
		Texture texture;
		texture.id = TextureFromFile(wide.c_str(), shader);
		texture.type = typeName;
		textures.push_back(texture);
	}
	return textures;
}


GLuint Model::TextureFromFile(const char* path, Shader* shader)
{

	//Generate texture ID and load texture data 
	GLuint tex;

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);


	int width, height;
	glActiveTexture(GL_TEXTURE0);
	

	unsigned char* image = SOIL_load_image(path, &width, &height, 0, SOIL_LOAD_RGB);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, image);
	
	SOIL_free_image_data(image);

	glUniform1i(glGetUniformLocation(shader->programHandle, "texture_diffuse"), 0);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	return tex;
}
