#include "ParticleSpawner.h"



struct Particle {
	glm::vec3 pos, speed;
	unsigned char r, g, b, a;
	float size, angle, weight, cameradistance;
	float life;

	bool operator<(const Particle& that)const
	{
		return this->cameradistance > that.cameradistance;
	}
};

Particle ParticlesContainer[200];
int ParticlesCount = 0;

CGUE17_SneakySteve::scene::ParticleSpawner::ParticleSpawner(Shader* shader, Camera* camera)
	:camera(camera), shader(shader)
{
	static const GLfloat vB_data[] = {
		-0.25f, -0.25f, 0.0f,
		0.25f, -0.25f, 0.0f,
		-0.25f, 0.25f, 0.0f,
		0.25f, 0.25f, 0.0f,
	};
	position = glm::vec3(4.0f, 2.0f, 6.0f);


	glGenVertexArrays(1, &this->VAO);


	glBindVertexArray(this->VAO);
	
	glGenBuffers(1, &particles_vertex_VBO);
	glBindBuffer(GL_ARRAY_BUFFER, particles_vertex_VBO);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vB_data), vB_data, GL_STATIC_DRAW);

	
	glGenBuffers(1, &particles_position_VBO);
	glBindBuffer(GL_ARRAY_BUFFER, particles_position_VBO);
	glBufferData(GL_ARRAY_BUFFER, MaxParticles * 4 * sizeof(GLfloat), NULL, GL_STREAM_DRAW);

	
	glGenBuffers(1, &particles_color_VBO);
	glBindBuffer(GL_ARRAY_BUFFER, particles_color_VBO);
	glBufferData(GL_ARRAY_BUFFER, MaxParticles * 4 * sizeof(GLubyte), NULL, GL_STREAM_DRAW);

	glBindVertexArray(0);

	for (int i = 0; i<MaxParticles; i++) {
		ParticlesContainer[i].life = -1.0f;
		ParticlesContainer[i].cameradistance = -1.0f;
	}
}

void CGUE17_SneakySteve::scene::ParticleSpawner::update(float deltaT)
{
	int newParticles = (int)(deltaT*200.0);
	if (newParticles > 50)
		newParticles = 50;

	for (int i = 0; i < newParticles; i++)
	{
		int particleIndex = FindUnusedParticle();
		ParticlesContainer[particleIndex].life = 1.0f;
		ParticlesContainer[particleIndex].pos = position;

		float spread = 1.5f;
		glm::vec3 maindir = glm::vec3(0.0f, rand() % 2, 0.0f);

		glm::vec3 randomdir = glm::vec3(
			(rand() % 2000 - 1000.0f) / 1000.0f,
			(rand() % 2000 - 1000.0f) / 1000.0f,
			(rand() % 2000 - 1000.0f) / 1000.0f
		);

		ParticlesContainer[particleIndex].speed = maindir + randomdir*spread;
		ParticlesContainer[particleIndex].r = rand() % 256;
		ParticlesContainer[particleIndex].g = rand() % 50;
		ParticlesContainer[particleIndex].b = rand() % 50;
		ParticlesContainer[particleIndex].a = (rand() % 256) / 3;

		ParticlesContainer[particleIndex].size = (rand() % 1000) / 2000.0f + 0.1f;
	}

	ParticlesCount = 0;
	for (int i = 0; i < MaxParticles; i++)
	{
		Particle& p = ParticlesContainer[i];

		if (p.life > 0.0f)
		{
			p.life -= deltaT;
			if (p.life > 0.0f)
			{
				p.speed += glm::vec3(0.0f, -1.0f, 0.0f) * (float)deltaT * 0.5f;
				p.pos += p.speed * (float)deltaT;
				p.cameradistance = (p.pos - camera->position).length();

				g_particule_position_size_data[4 * ParticlesCount + 0] = p.pos.x;
				g_particule_position_size_data[4 * ParticlesCount + 1] = p.pos.y;
				g_particule_position_size_data[4 * ParticlesCount + 2] = p.pos.z;

				g_particule_position_size_data[4 * ParticlesCount + 3] = p.size;

				g_particule_color_data[4 * ParticlesCount + 0] = p.r;
				g_particule_color_data[4 * ParticlesCount + 1] = p.g;
				g_particule_color_data[4 * ParticlesCount + 2] = p.b;
				g_particule_color_data[4 * ParticlesCount + 3] = p.a;
			}
			else
			{
				p.cameradistance = -1.0f;
			}
			ParticlesCount++;
		}
	}
	SortParticles();

	glBindVertexArray(this ->VAO);

	glBindBuffer(GL_ARRAY_BUFFER, particles_position_VBO);
	glBufferData(GL_ARRAY_BUFFER, MaxParticles * 4 * sizeof(GLfloat), NULL, GL_STREAM_DRAW);
	glBufferSubData(GL_ARRAY_BUFFER, 0, ParticlesCount * sizeof(GLfloat) * 4, g_particule_position_size_data);

	glBindBuffer(GL_ARRAY_BUFFER, particles_color_VBO);
	glBufferData(GL_ARRAY_BUFFER, MaxParticles * 4 * sizeof(GLubyte), NULL, GL_STREAM_DRAW);
	glBufferSubData(GL_ARRAY_BUFFER, 0, ParticlesCount * sizeof(GLubyte) * 4, g_particule_color_data);

	glBindVertexArray(0);
}

void CGUE17_SneakySteve::scene::ParticleSpawner::Draw()
{
	GLuint CameraRight_worldspace_ID = glGetUniformLocation(shader->programHandle, "CameraRight_worldspace");
	GLuint CameraUp_worldspace_ID = glGetUniformLocation(shader->programHandle, "CameraUp_worldspace");
	GLuint ViewProjMatrixID = glGetUniformLocation(shader->programHandle, "VP");

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	

	glUseProgram(shader->programHandle);

	glm::mat4 ViewMatrix = camera->getViewMatrix();
	glm::mat4 ViewProjectionMatrix = camera->getCameraMatrix();

	glUniform3f(CameraRight_worldspace_ID, ViewMatrix[0][0], ViewMatrix[1][0], ViewMatrix[2][0]);
	glUniform3f(CameraUp_worldspace_ID, ViewMatrix[0][1], ViewMatrix[1][1], ViewMatrix[2][1]);
	glUniformMatrix4fv(ViewProjMatrixID, 1, GL_FALSE, &ViewProjectionMatrix[0][0]);

	glBindVertexArray(this->VAO);

	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, particles_vertex_VBO);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);

	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, particles_position_VBO);
	glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, 0, (void*)0);

	glEnableVertexAttribArray(2);
	glBindBuffer(GL_ARRAY_BUFFER, particles_color_VBO);
	glVertexAttribPointer(2, 4, GL_UNSIGNED_BYTE, GL_TRUE, 0, (void*)0);

	glVertexAttribDivisor(0, 0);
	glVertexAttribDivisor(1, 1);
	glVertexAttribDivisor(2, 1);

	glDrawArraysInstanced(GL_TRIANGLE_STRIP, 0, 4, ParticlesCount);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);
	glDisableVertexAttribArray(2);

	glBindVertexArray(0);
	glDisable(GL_BLEND);
}

int CGUE17_SneakySteve::scene::ParticleSpawner::FindUnusedParticle()
{
	for (int i = LastUsedParticle; i<MaxParticles; i++) {
		if (ParticlesContainer[i].life < 0) {
			LastUsedParticle = i;
			return i;
		}
	}

	for (int i = 0; i<LastUsedParticle; i++) {
		if (ParticlesContainer[i].life < 0) {
			LastUsedParticle = i;
			return i;
		}
	}

	return 0; // All particles are taken, override the first one
}

void CGUE17_SneakySteve::scene::ParticleSpawner::SortParticles()
{
	std::sort(&ParticlesContainer[0], &ParticlesContainer[MaxParticles]);
}
