#version 330

in vec3 Normal;
in vec3 FragPos;
in vec4 shadowCoord;


layout (location = 0) out vec4 fragColor;
layout (location = 1) out vec4 bloomColor;

uniform vec3 viewPos;
uniform sampler2DShadow shadowMap;
uniform vec3 color;

struct Light{
vec3 position;
vec3 direction;
float fov;
};

uniform Light light;


void main()
{

	
	vec3 visColor = vec3(1,1,1);
	//AMBIENT
	float ambientStrength = 0.15f;
    vec3 ambient = ambientStrength * vec3(1,1,1);
	vec3 diffuse = vec3(0,0,0);
	vec3 specular = vec3(0,0,0);

	vec3 lightDir = normalize(light.position - FragPos);
	float theta = dot(lightDir, normalize(-light.direction)); 
	if(theta > light.fov)
	{
		vec3 norm = normalize(Normal);
	
		float diff = max(dot(norm,lightDir),0.1);

		theta = max(theta -3*(1 - theta),0);

		diff = theta*diff;
		//vec3 diffuse = (diff * Normal)+vec3(0.15,0.15,0.15);
		diffuse = diff * vec3(1,1,1);

		//SPECULAR
		float specularStrength = 0.5f;
		vec3 viewDir = normalize(viewPos - FragPos);
		vec3 reflectDir = reflect(-lightDir, norm);  
		float spec = pow(max(dot(viewDir, reflectDir), 0.0), 32);
		specular = specularStrength * spec * vec3(1,1,1);
		

		if(texture(shadowMap,vec3(shadowCoord.xyz)/shadowCoord.w) > ((shadowCoord.z)/shadowCoord.w))
		{
			visColor = color;
		}
	}
	
	float visibility = 1.0f;
	if(textureProj(shadowMap,shadowCoord) < ((shadowCoord.z)/shadowCoord.w))
	{
	visibility = 0.0f;
	}
	//float visibility =  texture( shadowMap, vec3(shadowCoord.xy, (shadowCoord.z)/shadowCoord.w) );

	

	visibility = textureProj(shadowMap,shadowCoord);


	vec3 lighting = ambient + visibility * diffuse + visibility * specular;

	fragColor =  vec4((lighting * visColor),1);
	//fragColor = vec4(lighting,1);

	if((lighting * visColor) == ambient)
		fragColor = vec4(0.1,0.1,0.1,1);

	if(color == vec3(0,2,0))
		fragColor = vec4(0.5,0.5,0.5,1);


	//Cell Shading
	fragColor.x = fragColor.x -(mod(fragColor.x, 0.1));
	fragColor.y = fragColor.y -(mod(fragColor.y, 0.1));
	fragColor.z = fragColor.z -(mod(fragColor.z, 0.1));

	if((fragColor.x > 0.5 || fragColor.y > 0.5 || fragColor.z > 0.5) && !((fragColor.x - fragColor.y <0.05)&&(fragColor.x -fragColor.z < 0.05)))
	{
		bloomColor = vec4(fragColor);
	}
	else
	bloomColor = vec4(0.1,0.1,0.1,0.0);

}