#version 330

in vec3 Normal;
in vec3 FragPos;

out vec3 color;

uniform vec3 lightPos;
uniform vec3 viewPos;

uniform vec3 position;
uniform vec3 direction;
uniform vec3 ambient;
uniform vec3 diffuse;
uniform vec3 specular;
uniform float constant;
uniform float linear;
uniform float quadratic;
uniform float cutOff;
uniform float outerCutOff;

void main()
{
	vec3 ambient_final = ambient;

	vec3 norm = normalize(Normal);
    vec3 lightDir = normalize(position - FragPos);
    float diff = max(dot(norm, lightDir), 0.1);
	vec3 diffuse_final = diffuse * diff;
    
	vec3 viewDir = normalize(viewPos - FragPos);
    vec3 reflectDir = reflect(-lightDir, norm);
    float spec = max(dot(viewDir, reflectDir), 0.0);
	vec3 specular_final = specular * spec;

	float theta = dot(lightDir, normalize(-direction));
    float epsilon = (cutOff - outerCutOff);
    float intensity = clamp((theta - outerCutOff) / epsilon, 0.0, 1.0);
    diffuse_final  *= intensity;
	specular_final *= intensity;

	float distance = length(position - FragPos);
    float attenuation = 1.0f / (constant + linear * distance + quadratic * (distance * distance));
    
	ambient_final  *= attenuation;
    diffuse_final  *= attenuation;
	specular_final *= attenuation;

	color = vec3(ambient_final+diffuse_final+specular_final);
}