#include <iostream>
#include <memory>
#include <sstream>
#include <stdlib.h>
#include <stdio.h>
#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include "Recources/Shader.hpp"
#include "Scene\SceneObject.hpp"
#include "Scene\Camera.hpp"
#include "Scene\Model.hpp"
#include "GameObjects\GameObj.hpp"
#include "Util\Shaderholder.h"
#include "GameObjects\Safe.h"

#include <btBulletDynamicsCommon.h>
#include <BulletCollision\Gimpact\btGImpactCollisionAlgorithm.h>
#include "PhysicsManager.hpp"
#include "Util\LvlLoader.hpp"
#include "GameObjects\Player.hpp"
#include "Util\BulletDebugDrawer.h"

#include "irrklang\irrKlang.h"
#include <ctime>
#include "Scene\ParticleSpawner.h"

#include "Util\FrustumCulling.hpp"
#include "Util\KeyMapping.hpp"

using namespace CGUE17_SneakySteve::scene;
using namespace sneakystv;
using namespace irrklang;

//#pragma comment(lib, "irrKlang.lib")

std::unique_ptr<SceneObject> cube;

static void APIENTRY DebugCallback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, const GLvoid* userParam);
static std::string FormatDebugOutput(GLenum source, GLenum type, GLuint id, GLenum severity, const char* msg);
void init(GLFWwindow* window);
void update(double deltaT);
void draw();
void cleanup();
void computeMatricesFromInputs(GLFWwindow* window, double deltaT);
void loadLvl(int lvl);
void atCollision();
void renderToTexture();
void bloom();
void renderQuad();

GLFWwindow* window;

std::vector<Shader*> shaders;
//std::vector<std::unique_ptr<Shader>> shaders;
//std::unique_ptr<Shader> spotlight;
std::unique_ptr<Camera> camera;
std::vector<std::vector<std::unique_ptr <GameObj>> > worldobjs;
std::vector<GameObj*> enemys;
Safe* safe;
Player* player;
GLuint framebuffer;
GLuint depthTexture;

ParticleSpawner* particle;

GLuint bloomFBO;
GLuint colorBuffers[2];
GLuint pingpongFBO[2];
GLuint pingpongColorbuffers[2];


//TODO: put in its own class
btBroadphaseInterface* broadphase;
btDefaultCollisionConfiguration* collisionConfiguration;
btCollisionDispatcher* dispatcher;
btSequentialImpulseConstraintSolver* solver;
btDiscreteDynamicsWorld* dynamicsWorld;

PhysicsManager* physicsmanager;

BulletDebugDrawer* debugDrawer;

ISoundEngine* se;

struct Light {
	glm::vec4 position;
	glm::vec3 ambient;
	glm::vec3 diffuse;
	glm::vec3 specular;
	glm::vec3 direction;
	float cutOff;
	float outerCutOff;
	float constant;
	float linear;
	float quadratic;
};
std::vector<Light> lights;

std::unique_ptr<FrustumCulling> frustumCulling;

int main(int argc, char** argv)
{
	if (!glfwInit())
	{
		printf("ERROR: Could not init glfw\n");
		system("PAUSE");
		exit(EXIT_FAILURE);
	}


	int width = 1280;
	int height = 720;	
	bool fullscreen = false;

	if (argc >= 3)
	{
		if ((std::stringstream(argv[1]) >> width).fail())
		{
			std::cerr << "ERROR: Could not parse first int";
			system("PAUSE");
			exit(EXIT_FAILURE);
		}
		if ((std::stringstream(argv[2]) >> height).fail())
		{
			std::cerr << "ERROR: Could not parse second int";
			system("PAUSE");
			exit(EXIT_FAILURE);
		}

		if (argc >= 4)
		{
			if ((std::stringstream(argv[3]) >> fullscreen).fail())
			{
				std::cerr << "ERROR: Could not parse 3rd int";
				system("PAUSE");
				exit(EXIT_FAILURE);
			}

		}
	}

	//Anti Aliasing (If weak Pc delete following line)
	glfwWindowHint(GLFW_SAMPLES, 4);

	//GLFW Setup
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);

	GLFWmonitor* monitor = nullptr;

	if (fullscreen)
	{
		monitor = glfwGetPrimaryMonitor();

		auto refresh_rate = 60;
		glfwWindowHint(GLFW_REFRESH_RATE, refresh_rate);
	}

	window = glfwCreateWindow(width, height, "Sneaky Steve", monitor, nullptr);

	if (!window)
	{
		printf("ERROR: Could not open window\n");
		glfwTerminate();
		system("PAUSE");
		exit(EXIT_FAILURE);
	}

	//Hide mouse (no more anoying cursor)
	glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
	glfwMakeContextCurrent(window);

	glewExperimental = true;

	if (glewInit() != GLEW_OK)
	{
		printf("ERROR: Could not init glew\n");
		system("PAUSE");
		exit(EXIT_FAILURE);
	}

	glDebugMessageCallback(DebugCallback, nullptr);
	glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);

	//physicsmanager = new PhysicsManager();
	se = createIrrKlangDevice();
	if (!se)
	{
		cout << "Error: Sound Engine could not be created" << endl;
	}
	se->play2D("Recources/Sound/spymusic.mp3");

	init(window);

	glClearColor(0.1f, 0.1f, 0.1f, 0.1f);
	glViewport(0, 0, width, height);

	auto time = glfwGetTime();

	debugDrawer->ToggleDebugFlag(btIDebugDraw::DBG_DrawWireframe);

	glEnable(GL_POLYGON_OFFSET_FILL);
	// Gameloop

	Shaderholder::getShader(SHADER_BLOOM)->useShader();
	GLint sceneLoc = glGetUniformLocation(Shaderholder::getShader(SHADER_BLOOM)->programHandle, "scene");
	glUniform1i(sceneLoc, 0);
	GLint bloomBlurLoc = glGetUniformLocation(Shaderholder::getShader(SHADER_BLOOM)->programHandle, "bloomBlur");
	glUniform1i(bloomBlurLoc, 1);
	
	while (!glfwWindowShouldClose(window))
	{
		//If ESC is pressed close window after this loop
		if (glfwGetKey(window, GLFW_KEY_ESCAPE))
			glfwSetWindowShouldClose(window, true);

		//Calc time since last loop
		auto time_new = glfwGetTime();
		auto time_delta = (float)(time_new - time);
		if (KeyMapping::getFrameTime())
			std::cout << "Frame Time: " << time_delta << std::endl;
		time = time_new;
		//printf("FPS : %f \n", (1/time_delta));

		//clear the WindowColor and Depth information for new drawing
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		//TODO: make it completely frame independent in case of really low fps.
		//physics step
		dynamicsWorld->stepSimulation(time_delta,1);


		atCollision();
		
		//Compute input
		computeMatricesFromInputs(window, time_delta);


		update(time_delta);


		draw();

		//put drawn canvas into window
		glfwSwapBuffers(window);

		//catch error msg and stuff
		glfwPollEvents();

		if (glGetError() != GL_NO_ERROR)
			printf("GL Error!");

	}
	//end gameloop
	
	cleanup();

	glfwTerminate();
	
	return EXIT_SUCCESS;
}

void init(GLFWwindow* window) 
{
	//TODO: move in own class
	//necessary bullet stuff for physics
	broadphase = new btDbvtBroadphase();
	collisionConfiguration = new btDefaultCollisionConfiguration();
	dispatcher = new btCollisionDispatcher(collisionConfiguration);

	btGImpactCollisionAlgorithm::registerAlgorithm(dispatcher);
	solver = new btSequentialImpulseConstraintSolver;
	dynamicsWorld = new btDiscreteDynamicsWorld(dispatcher, broadphase, solver, collisionConfiguration);
	dynamicsWorld->setGravity(btVector3(0, -10, 0));

	renderToTexture();
	

	//Camera setup
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glEnable(GL_CULL_FACE);
	int width;
	int height;
	glfwGetWindowSize(window, &width, &height);
	camera = std::make_unique<Camera>(glm::vec3(0, 10, 50),45.0f,height,width,1.0f,500.0f);
	Shaderholder::init();
	FrustumCulling::init();
	KeyMapping::init();
	//Load shaders here
	
	Shaderholder::addShader(SHADER_WORLD, std::make_unique<Shader>("Shader/mat_vert.vert", "Shader/mat_frag.frag"));
	Shaderholder::addShader(SHADER_TEX, std::make_unique<Shader>("Shader/texture_vert.vert", "Shader/texture_frag.frag"));
	Shaderholder::addShader(SHADER_BULLETDB, std::make_unique<Shader>("Shader/test_vertex.vert", "Shader/test_fragment.frag"));
	Shaderholder::addShader(SHADER_SHADOW, std::make_unique<Shader>("Shader/shadowMap.vert", "Shader/shadowMap.frag"));
	Shaderholder::addShader(SHADER_SPOTLIGHT, std::make_unique<Shader>("Shader/spot_light.vert", "Shader/spot_light.frag"));
	Shaderholder::addShader(SHADER_PARTICLE, std::make_unique<Shader>("Shader/particle.vert", "Shader/particle.frag"));
	Shaderholder::addShader(SHADER_GAUS, std::make_unique<Shader>("Shader/blur.vert", "Shader/blur.frag"));
	Shaderholder::addShader(SHADER_BLOOM, std::make_unique<Shader>("Shader/postBloom.vert", "Shader/postBloom.frag"));



	shaders.push_back(Shaderholder::getShader(SHADER_WORLD));
	shaders.push_back(Shaderholder::getShader(SHADER_TEX));
	shaders.push_back(Shaderholder::getShader(SHADER_BULLETDB));
	shaders.push_back(Shaderholder::getShader(SHADER_SHADOW));

	//Set size of worldobject shaders (needed to draw different Shaders)
	// 1st vector is to deiffer between shaders
	// 2nd vector saves the Object in the shaderlist
	worldobjs.resize(shaders.size());

	debugDrawer = new BulletDebugDrawer();
	debugDrawer->setDebugMode(btIDebugDraw::DBG_DrawWireframe);
	dynamicsWorld->setDebugDrawer(debugDrawer);

	bloom();

	//Load lvl parameter
	loadLvl(1);

}

void update(double deltaT)
{
	for (auto& shaderobjs : worldobjs)
	{
		for (auto& obj : shaderobjs)
		{
			obj->update(deltaT);
		}
		
	}
}

void draw()
{	
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(4.0f, 0.0f);
	
	glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
	glViewport(0, 0, 2048, 2048);
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
	glm::vec3 lightpos = player->position;
	lightpos.y = lightpos.y + 2;
	glm::vec3 lightInvDir = (camera->getCamFront() * -1.0f);
	lightInvDir.y = 0.0f;
	
	shaders.at(3)->useShader();

	glm::mat4 depthProjectionMatrix = glm::perspective<float>(60.0f, 1.0f, 2.0f, 100.0f);
	glm::mat4 depthViewMatrix = glm::lookAt(lightpos, lightpos-lightInvDir, glm::vec3(0, 1, 0));
	glm::mat4 depthModelMatrix = glm::mat4(1.0);
	glm::mat4 depthMVP = depthProjectionMatrix * depthViewMatrix * depthModelMatrix;

	auto depthMatrixID = glGetUniformLocation(shaders.at(3)->programHandle, "depthMVP");
	glUniformMatrix4fv(depthMatrixID, 1, GL_FALSE, &depthMVP[0][0]);

	glPolygonOffset(4,0);
	for (int i = 0; i < worldobjs.size(); i++)
	{
		for (auto& worldobj : worldobjs.at(i))
		{
			if (worldobj->rigidBody != player->rigidBody)
			{
				worldobj->draw(Shaderholder::getShader(SHADER_SHADOW), camera->getCameraMatrix());
			}
		}
	}

	glBindFramebuffer(GL_FRAMEBUFFER, bloomFBO);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPolygonOffset(0, 0);
	int width;
	int height;
	glfwGetWindowSize(window, &width, &height);
	glViewport(0, 0, width, height);
	//draw one Shader after the other.
	int culledObjects = 0;
	for (int i = 0; i < worldobjs.size(); i++)
	{
		shaders.at(i)->useShader();
		if (i == 0) 
		{
			auto depthBiasID = glGetUniformLocation(shaders.at(i)->programHandle, "depthBiasMVP");
			glm::mat4 biasMatrix(
			0.5, 0.0, 0.0, 0.0,
			0.0, 0.5, 0.0, 0.0,
			0.0, 0.0, 0.5, 0.0,
			0.5, 0.5, 0.5, 1.0
			);

			glm::mat4 depthBiasMVP = biasMatrix*depthMVP;
			glUniformMatrix4fv(depthBiasID, 1, GL_FALSE, &depthBiasMVP[0][0]);

			auto shadowMapID = glGetUniformLocation(shaders.at(i)->programHandle, "shadowMap");
			glActiveTexture(GL_TEXTURE0);
			glBindTexture(GL_TEXTURE_2D, depthTexture);
			glUniform1i(shadowMapID, 0);
		}

		//set Uniform Projection and viewmatrix for shader
		glm::mat4 view_projection = camera->getCameraMatrix();
		auto view_projection_location = glGetUniformLocation(shaders.at(i)->programHandle, "proj");
		glUniformMatrix4fv(view_projection_location, 1, GL_FALSE, glm::value_ptr(view_projection));

		//FrustumCulling::computePlanes((float*)glm::value_ptr(camera->getProjectionMatrix()*camera->getViewMatrix()));
		//FrustumCulling::computePlanes(glm::value_ptr(view_projection));
		//std::cout << FrustumCulling::planes[0]->normal.x << std::endl;

		//set Uniform Light for shader
		GLint light_position_location = glGetUniformLocation(shaders.at(i)->programHandle, "lightPos");
		glUniform3f(light_position_location, lightInvDir.x, lightInvDir.y, lightInvDir.z);

		//FOR SPECULAR
		GLint viewPosLoc = glGetUniformLocation(shaders.at(i)->programHandle, "viewPos");
		glUniform3f(viewPosLoc, camera->position.x, camera->position.y, camera->position.z);

		//Spotlight stuff
		GLint position = glGetUniformLocation(shaders.at(i)->programHandle, "light.position");
		glUniform3f(position, lightpos.x, lightpos.y, lightpos.z);
		GLint direction = glGetUniformLocation(shaders.at(i)->programHandle, "light.direction");
		glm::vec3 lightDir = lightInvDir * -1.0f;
		glUniform3f(direction, lightDir.x, lightDir.y, lightDir.z);
		float test = glm::cos(glm::radians(45.5f));
		glUniform1f(glGetUniformLocation(shaders.at(i)->programHandle, "light.fov"), glm::cos(glm::radians(45.5f)));
		
		
		/*
		glUniform1i(glGetUniformLocation(shaders.at(i)->programHandle, "material.diffuse"), 0);
		glUniform1i(glGetUniformLocation(shaders.at(i)->programHandle, "material.specular"), 1);
		glUniform1f(glGetUniformLocation(shaders.at(i)->programHandle, "material.shininess"), 32.0f);*/
		
		
		for (auto& worldobj : worldobjs.at(i))
		{
			culledObjects += worldobj->draw(shaders.at(i), camera->getCameraMatrix());
		}
		//auto time_d = static_cast<float>(glfwGetTime());
		//auto time_passed = 0 + time_d;
		//if (time_passed >= 1.0f)
		//{
		/*if(KeyMapping::getFrCulling())
			std::cout << "Culled Objects: " << culledObjects << std::endl;*/
			//time_passed = 0;
		//}

		
		
	}
	if (KeyMapping::getFrCulling())
		std::cout << "Culled Objects: " << culledObjects << std::endl;



	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	bool horizontal = true, first_iteration = true;
	GLuint amount = 2;
	Shaderholder::getShader(SHADER_GAUS)->useShader();

	for (unsigned int i = 0; i < amount; i++)
	{
		glBindFramebuffer(GL_FRAMEBUFFER, pingpongFBO[horizontal]);
		GLint horizontalLoc = glGetUniformLocation(Shaderholder::getShader(SHADER_GAUS)->programHandle, "horizontal");
		glUniform1i(horizontalLoc, horizontal);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, first_iteration ? colorBuffers[1] : pingpongColorbuffers[!horizontal]);
		renderQuad();
		horizontal = !horizontal;
		if (first_iteration)
			first_iteration = false;
	}
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	Shaderholder::getShader(SHADER_BLOOM)->useShader();
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, colorBuffers[0]);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, pingpongColorbuffers[!horizontal]);
	GLint bloomLoc = glGetUniformLocation(Shaderholder::getShader(SHADER_BLOOM)->programHandle, "bloom");
	glUniform1i(bloomLoc, KeyMapping::getBloom());
	renderQuad();


	if (debugDrawer->m_debugMode == btIDebugDraw::DBG_DrawWireframe) 
	{
		//Set Debug Shader
		dynamicsWorld->debugDrawWorld();
		Shaderholder::getShader(SHADER_BULLETDB)->useShader();
		glm::mat4 view_projection = camera->getCameraMatrix();
		auto view_projection_location = glGetUniformLocation(shaders.at(shaders.size() - 1)->programHandle, "proj");
		glUniformMatrix4fv(view_projection_location, 1, GL_FALSE, glm::value_ptr(view_projection));	

		//Get all the Lines in a vao
		debugDrawer->doDrawing();

		//Activate the vao
		glBindVertexArray(debugDrawer->vao);
		glDrawArrays(GL_LINES, 0, debugDrawer->LINES.size()*2);

		//Clean up
		debugDrawer->cleanDrawing();
	};	
}

void atCollision(/*std::vector<CGUE17_SneakySteve::scene::Mesh>& meshes*/) {

	struct ContactSensorCallback : public btCollisionWorld::ContactResultCallback {
		//! Constructor, pass whatever context you want to have available when processing contacts
		/*! You may also want to set m_collisionFilterGroup and m_collisionFilterMask
		*  (supplied by the superclass) for needsCollision() */

		ContactSensorCallback(btRigidBody* tgtBody, std::vector<GameObj*>* enemyList , GLFWwindow* window): btCollisionWorld::ContactResultCallback(), body(tgtBody), enemyList(enemyList), window(window){ }

		btRigidBody* body; //!< The body the sensor is monitoring
		std::vector<GameObj*>* enemyList;
		GLFWwindow* window;

		virtual bool needsCollision(btBroadphaseProxy* proxy) const {
			// superclass will check m_collisionFilterGroup and m_collisionFilterMask
			if (!btCollisionWorld::ContactResultCallback::needsCollision(proxy))
				return false;

			return body->checkCollideWithOverride(static_cast<btCollisionObject*>(proxy->m_clientObject));
		}

		virtual btScalar addSingleResult(btManifoldPoint& cp,
			const btCollisionObjectWrapper* colObj0, int partId0, int index0,
			const btCollisionObjectWrapper* colObj1, int partId1, int index1)
		{
			const btCollisionObjectWrapper* colObj;

			//Get the Object that is not the Player
			if (colObj0->m_collisionObject == body) 				
				colObj = colObj1;
			else 
				colObj = colObj0;

			//se->play2D("Recources/Sound/solid.wav", GL_FALSE);
			if (colObj->m_collisionObject == safe->rigidBody)
			{
				printf("YOU WIN\n");
				std::clock_t start;
				double duration = 0;
				start = std::clock();
				while (duration<2)
				{
					duration = (std::clock() - start) / (double)CLOCKS_PER_SEC;
				}
				glfwSetWindowShouldClose(window, true);
				
			}
			for (GameObj* enemy : enemys) {
				if (colObj->m_collisionObject == enemy->rigidBody) 
				{
					std::clock_t start;
					double duration = 0;
					start = std::clock();
					se->play2D("Recources/Sound/GameOver.mp3", GL_FALSE);
					while (/*se->isCurrentlyPlaying("GameOver.mp3")==0 &&*/ duration<2)
					{
						duration = (std::clock() - start) / (double)CLOCKS_PER_SEC;
					}
					glfwSetWindowShouldClose(window, true);
				}
			}
			return 0; 
		}
	};

	ContactSensorCallback callback(player->rigidBody, &enemys, window);
	dynamicsWorld->contactTest(player->rigidBody, callback);
}

void cleanup()
{
	cube.reset(nullptr);

	
	

	//Bullet stuff
	delete dynamicsWorld;
	delete solver;
	delete dispatcher;
	delete collisionConfiguration;
	delete broadphase;
	/*dynamicsWorld->removeRigidBody(fallRigidBody);
	delete fallRigidBody->getMotionState();
	delete fallRigidBody;
	dynamicsWorld->removeRigidBody(groundRigidBody);
	delete groundRigidBody->getMotionState();
	delete groundRigidBody;
	delete fallShape;
	delete groundShape;*/
	se->drop();
}

void computeMatricesFromInputs(GLFWwindow* window, double deltaT)
{
	float speed = 10.0f;
	double xpos, ypos;
	glfwGetCursorPos(window, &xpos, &ypos);

	camera->rotate(xpos, ypos);

	float xdir = 0.0f;
	float ydir = 0.0f;

	

	camera->translate(xdir, ydir, 0);
	//atCollision();
	
	float yRot = 0;

	if (glfwGetKey(window, GLFW_KEY_4))
		yRot += deltaT*speed;
	if (glfwGetKey(window, GLFW_KEY_6))
		yRot -= deltaT*speed;

	player->rotate(0,yRot);
	


	//TODO:make jump with z axis
	xdir = 0.0f;
	ydir = 0.0f;
	if (glfwGetKey(window, GLFW_KEY_SPACE)) 
	{
		player->jump();

		// play2D returns 0 if it is successfully playing
		string soundFile = "Recources/Sound/jump.mp3";
		if (se->play2D(soundFile.c_str(), GL_FALSE) != 0)
		{
			cout << "Error: Could not play file" << endl;
		}
		//se->play2D(soundFile.c_str(), GL_TRUE);
		// Keep playing until song/sound has ended
		//while (se->isCurrentlyPlaying(soundFile.c_str()))
		//	Sleep(100);
		//while (se->isCurrentlyPlaying("jump.mp3")==0 && duration<1)
		//{
		//	duration = (std::clock() - start) / (double)CLOCKS_PER_SEC;
		//}

	}

	if (glfwGetKey(window, GLFW_KEY_W))
		xdir += speed;
	if (glfwGetKey(window, GLFW_KEY_S))
		xdir -= speed;
	if (glfwGetKey(window, GLFW_KEY_A))
		ydir += speed;
	if (glfwGetKey(window, GLFW_KEY_D))
		ydir -= speed;
	

	player->translate(xdir, ydir, 0);

	if (glfwGetKey(window, GLFW_KEY_N))
	{
		debugDrawer->ToggleDebugFlag(btIDebugDraw::DBG_DrawWireframe);
	}
	if (glfwGetKey(window, GLFW_KEY_B)) {
		debugDrawer->ToggleDebugFlag(btIDebugDraw::DBG_DrawAabb);
		
	}
	
	//KeyMapping
	if (glfwGetKey(window, GLFW_KEY_F1))
	{
		std::cout << KeyMapping::setHelp() << std::endl;
	}
	if (glfwGetKey(window, GLFW_KEY_F2))
	{
		std::cout << KeyMapping::setFrameTime() << std::endl;
	}
	if (glfwGetKey(window, GLFW_KEY_F3))
	{
		std::cout << KeyMapping::setWireFrame() << std::endl;
	}
	if (glfwGetKey(window, GLFW_KEY_F6))
	{
		std::cout << KeyMapping::setParticle() << std::endl;
	}
	if (glfwGetKey(window, GLFW_KEY_F7))
	{
		std::cout << KeyMapping::setBloom() << std::endl;
	}
	if (glfwGetKey(window, GLFW_KEY_F8))
	{
		std::cout << KeyMapping::setFrCulling() << std::endl;
	}
	if (glfwGetKey(window, GLFW_KEY_F9))
	{
		std::cout << KeyMapping::setBlending() << std::endl;
	}


}

void loadLvl(int lvl)
{
	//make Player Gameobj
	btCollisionShape* fallShape = new btBoxShape(btVector3(0.5,0.5,0.5));
	btDefaultMotionState* fallMotionState = new btDefaultMotionState(btTransform(btQuaternion(0, 0, 0, 1), btVector3(0,5, 0)));
	btScalar mass = 50;
	btVector3 fallInertia(0, 0, 0);
	fallShape->calculateLocalInertia(mass, fallInertia);
	btRigidBody::btRigidBodyConstructionInfo fallRigidBodyCI(mass, fallMotionState, fallShape, fallInertia);
	btRigidBody* playerRigidBody = new btRigidBody(fallRigidBodyCI);
	playerRigidBody->setActivationState(DISABLE_DEACTIVATION);	//very fucking important!!!!!!!!!!!
	dynamicsWorld->addRigidBody(playerRigidBody);
	std::unique_ptr<Player> playerhelp = make_unique<Player>("Recources/Models/player.obj", shaders.at(0), glm::vec3(0, -0.25, 0), 0, 0, 0, playerRigidBody, camera.get());
	player = playerhelp.get();
	worldobjs.at(0).push_back(std::move(playerhelp));


	//make Safe
	btVector3 boxHalf = btVector3(5, 5, 2);
	fallShape = new btBoxShape(boxHalf);
	fallMotionState = new btDefaultMotionState(btTransform(btQuaternion(0,30.1, 0, 1), btVector3(-10, 2, -30.5)));
	fallRigidBodyCI = btRigidBody::btRigidBodyConstructionInfo(0, fallMotionState, fallShape, btVector3(0, 0, 0));
	btRigidBody* fallRigidBody = new btRigidBody(fallRigidBodyCI);
	dynamicsWorld->addRigidBody(fallRigidBody);
	std::unique_ptr<Safe> help = make_unique<Safe>("Recources/Models/vaultclosed.obj", Shaderholder::getShader(SHADER_WORLD), glm::vec3(0, 0, -0.1), 0, 0, 0, fallRigidBody);
	safe = help.get();
	worldobjs.at(0).push_back(std::move(help));

	LvlLoader::loadLvl(lvl, &shaders, &worldobjs, &enemys, dynamicsWorld, camera.get());
}

static void APIENTRY DebugCallback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, const GLvoid* userParam)
{
	std::string error = FormatDebugOutput(source, type, id, severity, message);
	std::cout << error << std::endl;
}

static std::string FormatDebugOutput(GLenum source, GLenum type, GLuint id, GLenum severity, const char* msg) {
	std::stringstream stringStream;
	std::string sourceString;
	std::string typeString;
	std::string severityString;

	// The AMD variant of this extension provides a less detailed classification of the error,
	// which is why some arguments might be "Unknown".
	switch (source) {
	case GL_DEBUG_CATEGORY_API_ERROR_AMD:
	case GL_DEBUG_SOURCE_API: {
		sourceString = "API";
		break;
	}
	case GL_DEBUG_CATEGORY_APPLICATION_AMD:
	case GL_DEBUG_SOURCE_APPLICATION: {
		sourceString = "Application";
		break;
	}
	case GL_DEBUG_CATEGORY_WINDOW_SYSTEM_AMD:
	case GL_DEBUG_SOURCE_WINDOW_SYSTEM: {
		sourceString = "Window System";
		break;
	}
	case GL_DEBUG_CATEGORY_SHADER_COMPILER_AMD:
	case GL_DEBUG_SOURCE_SHADER_COMPILER: {
		sourceString = "Shader Compiler";
		break;
	}
	case GL_DEBUG_SOURCE_THIRD_PARTY: {
		sourceString = "Third Party";
		break;
	}
	case GL_DEBUG_CATEGORY_OTHER_AMD:
	case GL_DEBUG_SOURCE_OTHER: {
		sourceString = "Other";
		break;
	}
	default: {
		sourceString = "Unknown";
		break;
	}
	}

	switch (type) {
	case GL_DEBUG_TYPE_ERROR: {
		typeString = "Error";
		break;
	}
	case GL_DEBUG_CATEGORY_DEPRECATION_AMD:
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR: {
		typeString = "Deprecated Behavior";
		break;
	}
	case GL_DEBUG_CATEGORY_UNDEFINED_BEHAVIOR_AMD:
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR: {
		typeString = "Undefined Behavior";
		break;
	}
	case GL_DEBUG_TYPE_PORTABILITY_ARB: {
		typeString = "Portability";
		break;
	}
	case GL_DEBUG_CATEGORY_PERFORMANCE_AMD:
	case GL_DEBUG_TYPE_PERFORMANCE: {
		typeString = "Performance";
		break;
	}
	case GL_DEBUG_CATEGORY_OTHER_AMD:
	case GL_DEBUG_TYPE_OTHER: {
		typeString = "Other";
		break;
	}
	default: {
		typeString = "Unknown";
		break;
	}
	}

	switch (severity) {
	case GL_DEBUG_SEVERITY_HIGH: {
		severityString = "High";
		break;
	}
	case GL_DEBUG_SEVERITY_MEDIUM: {
		severityString = "Medium";
		break;
	}
	case GL_DEBUG_SEVERITY_LOW: {
		severityString = "Low";
		break;
	}
	default: {
		severityString = "Unknown";
		break;
	}
	}

	stringStream << "OpenGL Error: " << msg;
	stringStream << " [Source = " << sourceString;
	stringStream << ", Type = " << typeString;
	stringStream << ", Severity = " << severityString;
	stringStream << ", ID = " << id << "]";

	return stringStream.str();
}

void renderToTexture()
{
	int width = 2048;
	int height = 2048;
	framebuffer = 0;
	glGenFramebuffers(1, &framebuffer);
	glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);

	glGenTextures(1, &depthTexture);
	glBindTexture(GL_TEXTURE_2D, depthTexture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT16, width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, 0);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LESS);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_REF_TO_TEXTURE);

	glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, depthTexture, 0);
	glDrawBuffer(GL_NONE);
	
	if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
		return;
}

void bloom()
{
	int width;
	int height;
	glfwGetWindowSize(window, &width, &height);
	glGenFramebuffers(1, &bloomFBO);
	glBindFramebuffer(GL_FRAMEBUFFER, bloomFBO);
	glGenTextures(2, colorBuffers);
	for (int i = 0; i < 2; i++)
	{
		glBindTexture(GL_TEXTURE_2D, colorBuffers[i]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F,width, height, 0, GL_RGB, GL_FLOAT, NULL);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		// attach texture to framebuffer
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + i, GL_TEXTURE_2D, colorBuffers[i], 0);
	}

	unsigned int rboDepth;
	glGenRenderbuffers(1, &rboDepth);
	glBindRenderbuffer(GL_RENDERBUFFER, rboDepth);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, rboDepth);

	unsigned int attachments[2] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1 };
	glDrawBuffers(2, attachments);

	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	glGenFramebuffers(2, pingpongFBO);
	glGenTextures(2, pingpongColorbuffers);
	for (unsigned int i = 0; i < 2; i++)
	{
		glBindFramebuffer(GL_FRAMEBUFFER, pingpongFBO[i]);
		glBindTexture(GL_TEXTURE_2D, pingpongColorbuffers[i]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F, width, height, 0, GL_RGB, GL_FLOAT, NULL);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE); // we clamp to the edge as the blur filter would otherwise sample repeated texture values!
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, pingpongColorbuffers[i], 0);
	}
}

unsigned int quadVAO = 0;
unsigned int quadVBO;
void renderQuad()
{
	if (quadVAO == 0)
	{
		float quadVertices[] = {
			// positions        // texture Coords
			-1.0f,  1.0f, 0.0f, 0.0f, 1.0f,
			-1.0f, -1.0f, 0.0f, 0.0f, 0.0f,
			1.0f,  1.0f, 0.0f, 1.0f, 1.0f,
			1.0f, -1.0f, 0.0f, 1.0f, 0.0f,
		};
		// setup plane VAO
		glGenVertexArrays(1, &quadVAO);
		glGenBuffers(1, &quadVBO);
		glBindVertexArray(quadVAO);
		glBindBuffer(GL_ARRAY_BUFFER, quadVBO);
		glBufferData(GL_ARRAY_BUFFER, sizeof(quadVertices), &quadVertices, GL_STATIC_DRAW);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)0);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
	}
	glBindVertexArray(quadVAO);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	glBindVertexArray(0);
}


